/*
 *  bignum_sub.c
 *  BigMath
 *
 *  Created by Curtis Jones on 2007.08.05.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "bignum.h"

// Knuth 4.3.1 - Algorithm S
//
// c - b = a
//
// minuend (c) - subtrahend (b) = difference (a)
//
bignum_t *
bignum_sub (bignum_t *dif, bignum_t *min, bignum_t *sub)
{
	u_int64_t j = 0;					// iterates through the magnitudes
	int       k = 0;					// tracks carries (-1, 0)
	u_int64_t b = dif->base;	// radix base
	
	__bignum_zero(dif);
	
	// if exactly one of the two numbers is negative,
	// perform addition after temporarily setting
	// that negative number to positive.
	//
	if (min->sign == BIGNUM_NEGATIVE && sub->sign == BIGNUM_POSITIVE) {
		min->sign = BIGNUM_POSITIVE;
		bignum_add(dif, min, sub);
		dif->sign = BIGNUM_NEGATIVE;
		min->sign = BIGNUM_NEGATIVE;
		return dif;
	}
	else if (min->sign == BIGNUM_POSITIVE && sub->sign == BIGNUM_NEGATIVE) {
		sub->sign = BIGNUM_POSITIVE;
		bignum_add(dif, min, sub);
		dif->sign = BIGNUM_POSITIVE;
		sub->sign = BIGNUM_NEGATIVE;
		return dif;
	}
	
	// cmp >  0 if min >  sub
	// cmp == 0 if min == sub
	// cmp <  0 if min <  sub
	//
	int cmp = bignum_cmp(min, sub);
	
	// if the minuend and the subtrahend are equal, then
	// the difference is equal to zero. set it and return.
	//
	if (0 == cmp)
		return dif;
	
	// if the minuend is less than the subtrahend, such that
	// the result of this operation will be negative, we need
	// to swap their positions, and set the sign of the 
	// difference to negative.
	//
	else if (0 > cmp) {
		bignum_sub(dif, sub, min);
		dif->sign = BIGNUM_NEGATIVE;
		return dif;
	}
	
	// we can assume that the minuend is greater than the
	// subtrahend at this point, so if the subtrahend is zero
	// then the difference is equal to the minuend. set it 
	// and return.
	//
	else if (bignum_iszero(sub)) {
		bignum_set_bn(dif, min);
		return dif;
	}
	
	// STEP S2
	//
	// sum_sub_j = (min_sub_j - sub_sub_j + k) mod b
	// k = floor_of(min_sub_j - sub_sub_j + k) / b
	//
	do {
		u_int64_t x, y;
		
		x = (u_int64_t)__bignum_get(min,j) - (u_int64_t)__bignum_get(sub,j) + k;
		k = x / b;
		y = x - (k * b);
		
		__bignum_set(dif, j, y);
		
		if (y != 0 && j >= dif->oofm)
			dif->oofm = j+1;
	}
	while (++j < min->oofm);
	
	return dif;
}
